#!/usr/bin/env python3 -B
# coding=utf-8

"""
VAIO Package Extractor
VAIO Packaging Manager Extractor
Copyright (C) 2019-2024 Plato Mavropoulos
"""

import os

from re import Match

from biosutilities.common.compression import is_szip_supported, szip_decompress
from biosutilities.common.paths import delete_file, make_dirs, path_name
from biosutilities.common.patterns import PAT_VAIO_CAB, PAT_VAIO_CFG, PAT_VAIO_CHK, PAT_VAIO_EXT
from biosutilities.common.system import printer
from biosutilities.common.templates import BIOSUtility


class VaioPackageExtract(BIOSUtility):
    """ VAIO Packaging Manager Extractor """

    TITLE: str = 'VAIO Packaging Manager Extractor'

    def check_format(self) -> bool:
        """ Check if input is VAIO Packaging Manager """

        return bool(PAT_VAIO_CFG.search(self.input_buffer))

    def parse_format(self) -> bool:
        """ Parse & Extract or Unlock VAIO Packaging Manager """

        input_name: str = path_name(self.input_object) if isinstance(self.input_object, str) else 'VAIO_Package'

        make_dirs(in_path=self.extract_path)

        if self._vaio_cabinet(name=input_name, buffer=self.input_buffer, extract_path=self.extract_path,
                              padding=self.padding) == 0:
            printer(message='Successfully Extracted!', padding=self.padding)
        elif self._vaio_unlock(name=input_name, buffer=self.input_buffer, extract_path=self.extract_path,
                               padding=self.padding) == 0:
            printer(message='Successfully Unlocked!', padding=self.padding)
        else:
            printer(message='Error: Failed to Extract or Unlock executable!', padding=self.padding)

            return False

        return True

    @staticmethod
    def _vaio_cabinet(name: str, buffer: bytes | bytearray, extract_path: str, padding: int = 0) -> int:
        """ Extract VAIO Packaging Manager executable """

        # Microsoft CAB Header XOR 0xFF
        match_cab: Match[bytes] | None = PAT_VAIO_CAB.search(buffer)

        if not match_cab:
            return 1

        printer(message='Detected obfuscated CAB archive!', padding=padding)

        # Get LE XOR CAB size
        cab_size: int = int.from_bytes(buffer[match_cab.start() + 0x8:match_cab.start() + 0xC], byteorder='little')

        # Create CAB size XOR value
        xor_size: int = int.from_bytes(b'\xFF' * 0x4, byteorder='little')

        # Perform XOR 0xFF and get actual CAB size
        cab_size ^= xor_size

        printer(message='Removing obfuscation...', padding=padding + 4)

        # Get BE XOR CAB data
        cab_data: int = int.from_bytes(buffer[match_cab.start():match_cab.start() + cab_size], byteorder='big')

        # Create CAB data XOR value
        xor_data: int = int.from_bytes(b'\xFF' * cab_size, byteorder='big')

        # Perform XOR 0xFF and get actual CAB data
        raw_data: bytes = (cab_data ^ xor_data).to_bytes(cab_size, 'big')

        printer(message='Extracting archive...', padding=padding + 4)

        cab_path: str = os.path.join(extract_path, f'{name}_Temporary.cab')

        # Create temporary CAB archive
        with open(cab_path, 'wb') as cab_file:
            cab_file.write(raw_data)

        if is_szip_supported(in_path=cab_path):
            if szip_decompress(in_path=cab_path, out_path=extract_path, in_name='VAIO CAB',
                               padding=padding + 8, check=True):
                delete_file(in_path=cab_path)
            else:
                return 3
        else:
            return 2

        return 0

    @staticmethod
    def _vaio_unlock(name: str, buffer: bytes | bytearray, extract_path: str, padding: int = 0) -> int:
        """ Unlock VAIO Packaging Manager executable """

        input_buffer: bytearray = bytearray(buffer) if isinstance(buffer, bytes) else buffer

        match_cfg: Match[bytes] | None = PAT_VAIO_CFG.search(input_buffer)

        if not match_cfg:
            return 1

        printer(message='Attempting to Unlock executable!', padding=padding)

        # Initialize VAIO Package Configuration file variables (assume overkill size of 0x500)
        cfg_bgn, cfg_end, cfg_false, cfg_true = [match_cfg.start(), match_cfg.start() + 0x500, b'', b'']

        # Get VAIO Package Configuration file info, split at new_line and stop at payload DOS header (EOF)
        cfg_info: list[bytearray] = input_buffer[cfg_bgn:cfg_end].split(
            b'\x0D\x0A\x4D\x5A')[0].replace(b'\x0D', b'').split(b'\x0A')

        printer(message='Retrieving True/False values...', padding=padding + 4)

        # Determine VAIO Package Configuration file True & False values
        for info in cfg_info:
            if info.startswith(b'ExtractPathByUser='):
                # Should be 0/No/False
                cfg_false = bytearray(b'0' if info[18:] in (b'0', b'1') else info[18:])

            if info.startswith(b'UseCompression='):
                # Should be 1/Yes/True
                cfg_true = bytearray(b'1' if info[15:] in (b'0', b'1') else info[15:])

        # Check if valid True/False values have been retrieved
        if cfg_false == cfg_true or not cfg_false or not cfg_true:
            printer(message='Error: Could not retrieve True/False values!', padding=padding + 8)

            return 2

        printer(message='Adjusting UseVAIOCheck entry...', padding=padding + 4)

        # Find and replace UseVAIOCheck entry from 1/Yes/True to 0/No/False
        vaio_check: Match[bytes] | None = PAT_VAIO_CHK.search(input_buffer[cfg_bgn:])

        if vaio_check:
            input_buffer[cfg_bgn + vaio_check.end():cfg_bgn + vaio_check.end() + len(cfg_true)] = cfg_false
        else:
            printer(message='Error: Could not find entry UseVAIOCheck!', padding=padding + 8)

            return 3

        printer(message='Adjusting ExtractPathByUser entry...', padding=padding + 4)

        # Find and replace ExtractPathByUser entry from 0/No/False to 1/Yes/True
        user_path: Match[bytes] | None = PAT_VAIO_EXT.search(input_buffer[cfg_bgn:])

        if user_path:
            input_buffer[cfg_bgn + user_path.end():cfg_bgn + user_path.end() + len(cfg_false)] = cfg_true
        else:
            printer(message='Error: Could not find entry ExtractPathByUser!', padding=padding + 8)

            return 4

        printer(message='Storing unlocked executable...', padding=padding + 4)

        # Store Unlocked VAIO Packaging Manager executable
        if vaio_check and user_path:
            unlock_path: str = os.path.join(extract_path, f'{name}_Unlocked.exe')

            with open(unlock_path, 'wb') as unl_file:
                unl_file.write(input_buffer)

        return 0



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Apple EFI IM4P
Apple EFI IM4P Splitter
Copyright (C) 2018-2024 Plato Mavropoulos
"""

import os

from re import Match
from typing import Final

from biosutilities.common.paths import is_file_read, make_dirs, path_size, path_stem
from biosutilities.common.patterns import PAT_APPLE_IM4P, PAT_INTEL_FD
from biosutilities.common.system import printer
from biosutilities.common.templates import BIOSUtility

from biosutilities.apple_efi_id import EFI_MAX_SIZE


class AppleEfiIm4pSplit(BIOSUtility):
    """ Apple EFI IM4P Splitter """

    TITLE: str = 'Apple EFI IM4P Splitter'

    # Intel Flash Descriptor Component Sizes (2MB, 4MB, 8MB, 16MB and 32MB)
    IFD_COMP_LEN: Final[dict[int, int]] = {2: 0x200000, 3: 0x400000, 4: 0x800000, 5: 0x1000000, 6: 0x2000000}

    def check_format(self) -> bool:
        """ Check if input is Apple EFI IM4P image """

        if isinstance(self.input_object, str) and is_file_read(in_path=self.input_object):
            if not self.input_object.lower().endswith('.im4p'):
                return False

            if path_size(in_path=self.input_object) > EFI_MAX_SIZE:
                return False

        if PAT_APPLE_IM4P.search(self.input_buffer) and PAT_INTEL_FD.search(self.input_buffer):
            return True

        return False

    def parse_format(self) -> bool:
        """ Parse & Split Apple EFI IM4P image """

        parse_success: bool = True

        make_dirs(in_path=self.extract_path)

        # Detect IM4P EFI pattern
        im4p_match: Match[bytes] | None = PAT_APPLE_IM4P.search(self.input_buffer)

        if not im4p_match:
            return False

        # After IM4P mefi (0x15), multi EFI payloads have _MEFIBIN (0x100) but is difficult to RE w/o varying samples.
        # However, _MEFIBIN is not required for splitting SPI images due to Intel Flash Descriptor Components Density.

        # IM4P mefi payload start offset
        mefi_data_bgn: int = im4p_match.start() + self.input_buffer[im4p_match.start() - 0x1]

        # IM4P mefi payload size
        mefi_data_len: int = int.from_bytes(self.input_buffer[im4p_match.end() + 0x5:im4p_match.end() + 0x9],
                                            byteorder='big')

        # Check if mefi is followed by _MEFIBIN
        mefibin_exist: bool = self.input_buffer[mefi_data_bgn:mefi_data_bgn + 0x8] == b'_MEFIBIN'

        # Actual multi EFI payloads start after _MEFIBIN
        efi_data_bgn: int = mefi_data_bgn + 0x100 if mefibin_exist else mefi_data_bgn

        # Actual multi EFI payloads size without _MEFIBIN
        efi_data_len: int = mefi_data_len - 0x100 if mefibin_exist else mefi_data_len

        # Adjust input file buffer to actual multi EFI payloads data
        input_buffer = self.input_buffer[efi_data_bgn:efi_data_bgn + efi_data_len]

        # Parse Intel Flash Descriptor pattern matches
        for ifd in PAT_INTEL_FD.finditer(input_buffer):
            # Component Base Address from FD start (ICH8-ICH10 = 1, IBX = 2, CPT+ = 3)
            ifd_flmap0_fcba: int = input_buffer[ifd.start() + 0x4] * 0x10

            # I/O Controller Hub (ICH)
            if ifd_flmap0_fcba == 0x10:
                # At ICH, Flash Descriptor starts at 0x0
                ifd_bgn_subtruct: int = 0x0

                # 0xBC for [0xAC] + 0xFF * 16 sanity check
                ifd_end_subtruct: int = 0xBC

            # Platform Controller Hub (PCH)
            else:
                # At PCH, Flash Descriptor starts at 0x10
                ifd_bgn_subtruct = 0x10

                # 0xBC for [0xAC] + 0xFF * 16 sanity check
                ifd_end_subtruct = 0xBC

            # Actual Flash Descriptor Start Offset
            ifd_match_start: int = ifd.start() - ifd_bgn_subtruct

            # Actual Flash Descriptor End Offset
            ifd_match_end: int = ifd.end() - ifd_end_subtruct

            # Calculate Intel Flash Descriptor Flash Component Total Size

            # Component Count (00 = 1, 01 = 2)
            ifd_flmap0_nc: int = ((int.from_bytes(input_buffer[ifd_match_end:ifd_match_end + 0x4],
                                                  byteorder='little') >> 8) & 3) + 1

            # PCH/ICH Strap Length (ME 2-8 & TXE 0-2 & SPS 1-2 <= 0x12, ME 9+ & TXE 3+ & SPS 3+ >= 0x13)
            ifd_flmap1_isl: int = input_buffer[ifd_match_end + 0x7]

            # Component Density Byte (ME 2-8 & TXE 0-2 & SPS 1-2 = 0:5, ME 9+ & TXE 3+ & SPS 3+ = 0:7)
            ifd_comp_den: int = input_buffer[ifd_match_start + ifd_flmap0_fcba]

            # Component 1 Density Bits (ME 2-8 & TXE 0-2 & SPS 1-2 = 3, ME 9+ & TXE 3+ & SPS 3+ = 4)
            ifd_comp_1_bitwise: int = 0xF if ifd_flmap1_isl >= 0x13 else 0x7

            # Component 2 Density Bits (ME 2-8 & TXE 0-2 & SPS 1-2 = 3, ME 9+ & TXE 3+ & SPS 3+ = 4)
            ifd_comp_2_bitwise: int = 0x4 if ifd_flmap1_isl >= 0x13 else 0x3

            # Component 1 Density (FCBA > C0DEN)
            ifd_comp_all_size: int = self.IFD_COMP_LEN[ifd_comp_den & ifd_comp_1_bitwise]

            # Component 2 Density (FCBA > C1DEN)
            if ifd_flmap0_nc == 2:
                ifd_comp_all_size += self.IFD_COMP_LEN[ifd_comp_den >> ifd_comp_2_bitwise]

            ifd_data_bgn: int = ifd_match_start
            ifd_data_end: int = ifd_data_bgn + ifd_comp_all_size

            ifd_data_txt: str = f'0x{ifd_data_bgn:07X}-0x{ifd_data_end:07X}'

            output_data: bytes = input_buffer[ifd_data_bgn:ifd_data_end]

            output_size: int = len(output_data)

            output_name: str = path_stem(in_path=self.input_object) if isinstance(self.input_object, str) else 'Part'

            output_path: str = os.path.join(self.extract_path, f'{output_name}_[{ifd_data_txt}].fd')

            with open(output_path, 'wb') as output_image:
                output_image.write(output_data)

            printer(message=f'Split Apple EFI image at {ifd_data_txt}!', padding=self.padding)

            if output_size != ifd_comp_all_size:
                printer(message=f'Error: Bad image size 0x{output_size:07X}, expected 0x{ifd_comp_all_size:07X}!',
                        padding=self.padding + 4)

                parse_success = False

        return parse_success



#!/usr/bin/env python3 -B
# coding=utf-8

"""
AMI UCP Extract
AMI UCP Update Extractor
Copyright (C) 2021-2024 Plato Mavropoulos
"""

import contextlib
import ctypes
import os
import re
import struct

from typing import Any, Final

from biosutilities.common.checksums import checksum_16
from biosutilities.common.compression import efi_decompress, is_efi_compressed
from biosutilities.common.paths import agnostic_path, delete_file, extract_folder, make_dirs, safe_name, safe_path
from biosutilities.common.patterns import PAT_AMI_UCP, PAT_INTEL_ENGINE
from biosutilities.common.structs import CHAR, ctypes_struct, UINT8, UINT16, UINT32
from biosutilities.common.system import printer
from biosutilities.common.templates import BIOSUtility
from biosutilities.common.texts import to_string

from biosutilities.ami_pfat_extract import AmiPfatExtract
from biosutilities.insyde_ifd_extract import InsydeIfdExtract


class UafHeader(ctypes.LittleEndianStructure):
    """ UAF Header """

    _pack_ = 1
    _fields_ = [
        ('ModuleTag',       CHAR * 4),      # 0x00
        ('ModuleSize',      UINT32),        # 0x04
        ('Checksum',        UINT16),        # 0x08
        ('Unknown0',        UINT8),         # 0x0A
        ('Unknown1',        UINT8),         # 0x0A
        ('Reserved',        UINT8 * 4)      # 0x0C
        # 0x10
    ]

    def _get_reserved(self) -> str:
        res_bytes: bytes = bytes(self.Reserved)

        res_hex: str = f'0x{int.from_bytes(res_bytes, byteorder="big"):0{0x4 * 2}X}'

        res_str: str = re.sub(r'[\n\t\r\x00 ]', '', res_bytes.decode('utf-8', 'ignore'))

        res_txt: str = f' ({res_str})' if len(res_str) else ''

        return f'{res_hex}{res_txt}'

    def get_tag(self) -> str:
        """ Get UAF Module Tag """

        return self.ModuleTag.decode('utf-8')

    def struct_print(self, padding: int = 0) -> None:
        """ Display structure information """

        printer(message=['Tag          :', self.get_tag()], padding=padding, new_line=False)
        printer(message=['Size         :', f'0x{self.ModuleSize:X}'], padding=padding, new_line=False)
        printer(message=['Checksum     :', f'0x{self.Checksum:04X}'], padding=padding, new_line=False)
        printer(message=['Unknown 0    :', f'0x{self.Unknown0:02X}'], padding=padding, new_line=False)
        printer(message=['Unknown 1    :', f'0x{self.Unknown1:02X}'], padding=padding, new_line=False)
        printer(message=['Reserved     :', self._get_reserved()], padding=padding, new_line=False)


class UafModule(ctypes.LittleEndianStructure):
    """ UAF Module """

    _pack_ = 1
    _fields_ = [
        ('CompressSize',    UINT32),        # 0x00
        ('OriginalSize',    UINT32)         # 0x04
        # 0x08
    ]

    def struct_print(self, filename: str, description: str, padding: int = 0) -> None:
        """ Display structure information """

        printer(message=['Compress Size:', f'0x{self.CompressSize:X}'], padding=padding, new_line=False)
        printer(message=['Original Size:', f'0x{self.OriginalSize:X}'], padding=padding, new_line=False)
        printer(message=['Filename     :', filename], padding=padding, new_line=False)
        printer(message=['Description  :', description], padding=padding, new_line=False)


class UiiHeader(ctypes.LittleEndianStructure):
    """ UII Header """

    _pack_ = 1
    _fields_ = [
        ('UIISize',         UINT16),        # 0x00
        ('Checksum',        UINT16),        # 0x02
        ('UtilityVersion',  UINT32),        # 0x04 AFU|BGT (Unknown, Signed)
        ('InfoSize',        UINT16),        # 0x08
        ('SupportBIOS',     UINT8),         # 0x0A
        ('SupportOS',       UINT8),         # 0x0B
        ('DataBusWidth',    UINT8),         # 0x0C
        ('ProgramType',     UINT8),         # 0x0D
        ('ProgramMode',     UINT8),         # 0x0E
        ('SourceSafeRel',   UINT8)          # 0x0F
        # 0x10
    ]

    SBI: Final[dict[int, str]] = {1: 'ALL', 2: 'AMIBIOS8', 3: 'UEFI', 4: 'AMIBIOS8/UEFI'}
    SOS: Final[dict[int, str]] = {1: 'DOS', 2: 'EFI', 3: 'Windows', 4: 'Linux', 5: 'FreeBSD',
                                  6: 'MacOS', 128: 'Multi-Platform'}
    DBW: Final[dict[int, str]] = {1: '16b', 2: '16/32b', 3: '32b', 4: '64b'}
    PTP: Final[dict[int, str]] = {1: 'Executable', 2: 'Library', 3: 'Driver'}
    PMD: Final[dict[int, str]] = {1: 'API', 2: 'Console', 3: 'GUI', 4: 'Console/GUI'}

    def struct_print(self, description: str, padding: int = 0) -> None:
        """ Display structure information """

        support_bios: str = self.SBI.get(self.SupportBIOS, f'Unknown ({self.SupportBIOS})')
        support_os: str = self.SOS.get(self.SupportOS, f'Unknown ({self.SupportOS})')
        data_bus_width: str = self.DBW.get(self.DataBusWidth, f'Unknown ({self.DataBusWidth})')
        program_type: str = self.PTP.get(self.ProgramType, f'Unknown ({self.ProgramType})')
        program_mode: str = self.PMD.get(self.ProgramMode, f'Unknown ({self.ProgramMode})')

        printer(message=['UII Size      :', f'0x{self.UIISize:X}'], padding=padding, new_line=False)
        printer(message=['Checksum      :', f'0x{self.Checksum:04X}'], padding=padding, new_line=False)
        printer(message=['Tool Version  :', f'0x{self.UtilityVersion:08X}'], padding=padding, new_line=False)
        printer(message=['Info Size     :', f'0x{self.InfoSize:X}'], padding=padding, new_line=False)
        printer(message=['Supported BIOS:', support_bios], padding=padding, new_line=False)
        printer(message=['Supported OS  :', support_os], padding=padding, new_line=False)
        printer(message=['Data Bus Width:', data_bus_width], padding=padding, new_line=False)
        printer(message=['Program Type  :', program_type], padding=padding, new_line=False)
        printer(message=['Program Mode  :', program_mode], padding=padding, new_line=False)
        printer(message=['SourceSafe Tag:', f'{self.SourceSafeRel:02d}'], padding=padding, new_line=False)
        printer(message=['Description   :', description], padding=padding, new_line=False)


class DisHeader(ctypes.LittleEndianStructure):
    """ DIS Header """

    _pack_ = 1
    _fields_ = [
        ('PasswordSize',    UINT16),        # 0x00
        ('EntryCount',      UINT16),        # 0x02
        ('Password',        CHAR * 12)      # 0x04
        # 0x10
    ]

    def struct_print(self, padding: int = 0) -> None:
        """ Display structure information """

        printer(message=['Password Size:', f'0x{self.PasswordSize:X}'], padding=padding, new_line=False)
        printer(message=['Entry Count  :', self.EntryCount], padding=padding, new_line=False)
        printer(message=['Password     :', self.Password.decode('utf-8')], padding=padding, new_line=False)


class DisModule(ctypes.LittleEndianStructure):
    """ DIS Module """

    _pack_ = 1
    _fields_ = [
        ('EnabledDisabled', UINT8),         # 0x00
        ('ShownHidden',     UINT8),         # 0x01
        ('Command',         CHAR * 32),     # 0x02
        ('Description',     CHAR * 256)     # 0x22
        # 0x122
    ]

    ENDIS: Final[dict[int, str]] = {0: 'Disabled', 1: 'Enabled'}
    SHOWN: Final[dict[int, str]] = {0: 'Hidden', 1: 'Shown', 2: 'Shown Only'}

    def struct_print(self, padding: int = 0) -> None:
        """ Display structure information """

        enabled_disabled: str = self.ENDIS.get(self.EnabledDisabled, f'Unknown ({self.EnabledDisabled})')
        shown_hidden: str = self.SHOWN.get(self.ShownHidden, f'Unknown ({self.ShownHidden})')
        command: str = self.Command.decode('utf-8').strip()
        description: str = self.Description.decode('utf-8').strip()

        printer(message=['State      :', enabled_disabled], padding=padding, new_line=False)
        printer(message=['Display    :', shown_hidden], padding=padding, new_line=False)
        printer(message=['Command    :', command], padding=padding, new_line=False)
        printer(message=['Description:', description], padding=padding, new_line=False)


class AmiUcpExtract(BIOSUtility):
    """ AMI UCP Update Extractor """

    TITLE: str = 'AMI UCP Update Extractor'

    # Get common ctypes Structure Sizes
    UAF_HDR_LEN: Final[int] = ctypes.sizeof(UafHeader)
    UAF_MOD_LEN: Final[int] = ctypes.sizeof(UafModule)
    DIS_HDR_LEN: Final[int] = ctypes.sizeof(DisHeader)
    DIS_MOD_LEN: Final[int] = ctypes.sizeof(DisModule)
    UII_HDR_LEN: Final[int] = ctypes.sizeof(UiiHeader)

    # AMI UCP Tag Dictionary
    UAF_TAG_DICT: Final[dict[str, list[str]]] = {
        '@3FI': ['HpBiosUpdate32.efi', 'HpBiosUpdate32.efi', ''],
        '@3S2': ['HpBiosUpdate32.s12', 'HpBiosUpdate32.s12', ''],
        '@3S4': ['HpBiosUpdate32.s14', 'HpBiosUpdate32.s14', ''],
        '@3S9': ['HpBiosUpdate32.s09', 'HpBiosUpdate32.s09', ''],
        '@3SG': ['HpBiosUpdate32.sig', 'HpBiosUpdate32.sig', ''],
        '@AMI': ['UCP_Nested.bin', 'Nested AMI UCP', ''],
        '@B12': ['BiosMgmt.s12', 'BiosMgmt.s12', ''],
        '@B14': ['BiosMgmt.s14', 'BiosMgmt.s14', ''],
        '@B32': ['BiosMgmt32.s12', 'BiosMgmt32.s12', ''],
        '@B34': ['BiosMgmt32.s14', 'BiosMgmt32.s14', ''],
        '@B39': ['BiosMgmt32.s09', 'BiosMgmt32.s09', ''],
        '@B3E': ['BiosMgmt32.efi', 'BiosMgmt32.efi', ''],
        '@BM9': ['BiosMgmt.s09', 'BiosMgmt.s09', ''],
        '@BME': ['BiosMgmt.efi', 'BiosMgmt.efi', ''],
        '@CKV': ['Check_Version.txt', 'Check Version', 'Text'],
        '@CMD': ['AFU_Command.txt', 'AMI AFU Command', 'Text'],
        '@CML': ['CMOSD4.txt', 'CMOS Item Number-Value (MSI)', 'Text'],
        '@CMS': ['CMOSD4.exe', 'Get or Set CMOS Item (MSI)', ''],
        '@CPM': ['AC_Message.txt', 'Confirm Power Message', ''],
        '@D32': ['amifldrv32.sys', 'amifldrv32.sys', ''],
        '@D64': ['amifldrv64.sys', 'amifldrv64.sys', ''],
        '@DCT': ['DevCon32.exe', 'Device Console WIN32', ''],
        '@DCX': ['DevCon64.exe', 'Device Console WIN64', ''],
        '@DFE': ['HpDevFwUpdate.efi', 'HpDevFwUpdate.efi', ''],
        '@DFS': ['HpDevFwUpdate.s12', 'HpDevFwUpdate.s12', ''],
        '@DIS': ['Command_Status.bin', 'Default Command Status', ''],
        '@ENB': ['ENBG64.exe', 'ENBG64.exe', ''],
        '@HPU': ['UCP_Main.bin', 'Utility Auxiliary File (HP)', ''],
        '@INS': ['Insyde_Nested.bin', 'Nested Insyde SFX', ''],
        '@M32': ['HpBiosMgmt32.s12', 'HpBiosMgmt32.s12', ''],
        '@M34': ['HpBiosMgmt32.s14', 'HpBiosMgmt32.s14', ''],
        '@M39': ['HpBiosMgmt32.s09', 'HpBiosMgmt32.s09', ''],
        '@M3I': ['HpBiosMgmt32.efi', 'HpBiosMgmt32.efi', ''],
        '@MEC': ['FWUpdLcl.txt', 'Intel FWUpdLcl Command', 'Text'],
        '@MED': ['FWUpdLcl_DOS.exe', 'Intel FWUpdLcl DOS', ''],
        '@MET': ['FWUpdLcl_WIN32.exe', 'Intel FWUpdLcl WIN32', ''],
        '@MFI': ['HpBiosMgmt.efi', 'HpBiosMgmt.efi', ''],
        '@MS2': ['HpBiosMgmt.s12', 'HpBiosMgmt.s12', ''],
        '@MS4': ['HpBiosMgmt.s14', 'HpBiosMgmt.s14', ''],
        '@MS9': ['HpBiosMgmt.s09', 'HpBiosMgmt.s09', ''],
        '@NAL': ['UCP_List.txt', 'AMI UCP Module Name List', ''],
        '@OKM': ['OK_Message.txt', 'OK Message', ''],
        '@PFC': ['BGT_Command.txt', 'AMI BGT Command', 'Text'],
        '@R3I': ['CryptRSA32.efi', 'CryptRSA32.efi', ''],
        '@RFI': ['CryptRSA.efi', 'CryptRSA.efi', ''],
        '@UAF': ['UCP_Main.bin', 'Utility Auxiliary File (AMI)', ''],
        '@UFI': ['HpBiosUpdate.efi', 'HpBiosUpdate.efi', ''],
        '@UII': ['UCP_Info.txt', 'Utility Identification Information', ''],
        '@US2': ['HpBiosUpdate.s12', 'HpBiosUpdate.s12', ''],
        '@US4': ['HpBiosUpdate.s14', 'HpBiosUpdate.s14', ''],
        '@US9': ['HpBiosUpdate.s09', 'HpBiosUpdate.s09', ''],
        '@USG': ['HpBiosUpdate.sig', 'HpBiosUpdate.sig', ''],
        '@VER': ['OEM_Version.txt', 'OEM Version', 'Text'],
        '@VXD': ['amifldrv.vxd', 'amifldrv.vxd', ''],
        '@W32': ['amifldrv32.sys', 'amifldrv32.sys', ''],
        '@W64': ['amifldrv64.sys', 'amifldrv64.sys', '']
    }

    def __init__(self, input_object: str | bytes | bytearray = b'', extract_path: str = '', padding: int = 0,
                 checksum: bool = False) -> None:
        super().__init__(input_object=input_object, extract_path=extract_path, padding=padding)

        self.checksum: bool = checksum

    def check_format(self) -> bool:
        """ Check if input is AMI UCP image """

        return bool(self._get_ami_ucp())

    def parse_format(self) -> bool:
        """ Parse & Extract AMI UCP structures """

        nal_dict: dict[str, tuple[str, str]] = {}  # Initialize @NAL Dictionary per UCP

        printer(message='Utility Configuration Program', padding=self.padding)

        make_dirs(in_path=self.extract_path)

        # Get best AMI UCP Pattern match based on @UAF|@HPU Size
        ucp_buffer: bytes = self._get_ami_ucp()

        # Parse @UAF|@HPU Header Structure
        uaf_hdr: Any = ctypes_struct(buffer=ucp_buffer, start_offset=0, class_object=UafHeader)

        ucp_tag: str = uaf_hdr.get_tag()

        printer(message=f'Utility Auxiliary File > {ucp_tag}:\n', padding=self.padding + 4)

        uaf_hdr.struct_print(padding=self.padding + 8)

        # Generate UafModule Structure
        fake: bytes = struct.pack('<II', len(ucp_buffer), len(ucp_buffer))

        # Parse @UAF|@HPU Module EFI Structure
        uaf_mod: Any = ctypes_struct(buffer=fake, start_offset=0, class_object=UafModule)

        # Get @UAF|@HPU Module Filename
        uaf_name: str = self.UAF_TAG_DICT[ucp_tag][0]

        # Get @UAF|@HPU Module Description
        uaf_desc: str = self.UAF_TAG_DICT[ucp_tag][1]

        # Print @UAF|@HPU Module EFI Info
        uaf_mod.struct_print(filename=uaf_name, description=uaf_desc, padding=self.padding + 8)

        if self.checksum:
            self._chk16_validate(data=ucp_buffer, tag=ucp_tag, padding=self.padding + 8)

        uaf_all: list[list] = self._get_uaf_mod(buffer=ucp_buffer, uaf_off=self.UAF_HDR_LEN)

        for mod_info in uaf_all:
            nal_dict = self._uaf_extract(buffer=ucp_buffer, extract_path=self.extract_path, mod_info=mod_info,
                                         nal_dict=nal_dict, padding=self.padding + 8)

        return True

    @staticmethod
    def _chk16_validate(data: bytes | bytearray, tag: str, padding: int = 0) -> None:
        """ Validate UCP Module Checksum-16 """

        if checksum_16(data=data) != 0:
            printer(message=f'Error: Invalid UCP Module {tag} Checksum!', padding=padding)
        else:
            printer(message=f'Checksum of UCP Module {tag} is valid!', padding=padding)

    def _get_ami_ucp(self) -> bytes:
        """ Get all input file AMI UCP patterns """

        uaf_mod_dat_max: bytes = b''

        uaf_mod_len_max: int = 0

        for uaf_match in PAT_AMI_UCP.finditer(self.input_buffer):
            uaf_mod_off: int = uaf_match.start()

            uaf_mod_buf: bytes = self.input_buffer[uaf_mod_off:]

            if len(uaf_mod_buf) <= self.UAF_HDR_LEN:
                continue

            uaf_hdr: Any = ctypes_struct(buffer=uaf_mod_buf, start_offset=0, class_object=UafHeader)

            uaf_mod_len: int = uaf_hdr.ModuleSize

            uaf_mod_dat: bytes = uaf_mod_buf[:uaf_mod_len]

            if uaf_mod_len != len(uaf_mod_dat):
                continue

            if checksum_16(data=uaf_mod_dat) != 0:
                continue

            if uaf_mod_len > uaf_mod_len_max:
                uaf_mod_len_max = uaf_mod_len

                uaf_mod_dat_max = uaf_mod_dat

        return uaf_mod_dat_max

    @staticmethod
    def _get_uaf_mod(buffer: bytes | bytearray, uaf_off: int = 0x0) -> list[list]:
        """ Get list of @UAF|@HPU Modules """

        uaf_all: list[list] = []  # Initialize list of all @UAF|@HPU Modules

        while buffer[uaf_off] == 0x40:  # ASCII of @ is 0x40
            # Parse @UAF|@HPU Module Structure
            uaf_hdr: Any = ctypes_struct(buffer=buffer, start_offset=uaf_off, class_object=UafHeader)

            uaf_tag: str = uaf_hdr.get_tag()  # Get unique @UAF|@HPU Module Tag

            uaf_all.append([uaf_tag, uaf_off, uaf_hdr])  # Store @UAF|@HPU Module Info

            uaf_off += uaf_hdr.ModuleSize  # Adjust to next @UAF|@HPU Module offset

            if uaf_off >= len(buffer):
                break  # Stop parsing at EOF

        # Check if @UAF|@HPU Module @NAL exists and place it first
        # Parsing @NAL first allows naming all @UAF|@HPU Modules
        for mod_idx, mod_val in enumerate(uaf_all):
            if mod_val[0] == '@NAL':
                uaf_all.insert(1, uaf_all.pop(mod_idx))  # After UII for visual purposes

                break  # @NAL found, skip the rest

        return uaf_all

    def _uaf_extract(self, buffer: bytes | bytearray, extract_path: str, mod_info: list,
                     nal_dict: dict[str, tuple[str, str]], padding: int = 0) -> dict[str, tuple[str, str]]:
        """ Parse & Extract AMI UCP > @UAF|@HPU Module/Section """

        uaf_tag: str = mod_info[0]
        uaf_off: int = mod_info[1]
        uaf_hdr: Any = mod_info[2]

        uaf_data_all: bytes = buffer[uaf_off:uaf_off + uaf_hdr.ModuleSize]  # @UAF|@HPU Module Entire Data

        uaf_data_mod: bytes = uaf_data_all[self.UAF_HDR_LEN:]  # @UAF|@HPU Module EFI Data

        uaf_data_raw: bytes = uaf_data_mod[self.UAF_MOD_LEN:]  # @UAF|@HPU Module Raw Data

        printer(message=f'Utility Auxiliary File > {uaf_tag}:\n', padding=padding)

        uaf_hdr.struct_print(padding=padding + 4)  # Print @UAF|@HPU Module Info

        # Parse UAF Module EFI Structure
        uaf_mod: Any = ctypes_struct(buffer=buffer, start_offset=uaf_off + self.UAF_HDR_LEN, class_object=UafModule)

        is_comp: bool = uaf_mod.CompressSize != uaf_mod.OriginalSize  # Detect @UAF|@HPU Module EFI Compression

        if uaf_tag in nal_dict:
            uaf_name: str = nal_dict[uaf_tag][1]  # Always prefer @NAL naming first
        elif uaf_tag in self.UAF_TAG_DICT:
            uaf_name = self.UAF_TAG_DICT[uaf_tag][0]  # Otherwise use built-in naming
        elif uaf_tag == '@ROM':
            uaf_name = 'BIOS.bin'  # BIOS/PFAT Firmware (w/o Signature)
        elif uaf_tag.startswith('@R0'):
            uaf_name = f'BIOS_0{uaf_tag[3:]}.bin'  # BIOS/PFAT Firmware
        elif uaf_tag.startswith('@S0'):
            uaf_name = f'BIOS_0{uaf_tag[3:]}.sig'  # BIOS/PFAT Signature
        elif uaf_tag.startswith('@DR'):
            uaf_name = f'DROM_0{uaf_tag[3:]}.bin'  # Thunderbolt Retimer Firmware
        elif uaf_tag.startswith('@DS'):
            uaf_name = f'DROM_0{uaf_tag[3:]}.sig'  # Thunderbolt Retimer Signature
        elif uaf_tag.startswith('@EC'):
            uaf_name = f'EC_0{uaf_tag[3:]}.bin'  # Embedded Controller Firmware
        elif uaf_tag.startswith('@ME'):
            uaf_name = f'ME_0{uaf_tag[3:]}.bin'  # Management Engine Firmware
        else:
            uaf_name = uaf_tag  # Could not name the @UAF|@HPU Module, use Tag instead

        uaf_fext: str = '' if uaf_name != uaf_tag else '.bin'

        uaf_fdesc: str = self.UAF_TAG_DICT[uaf_tag][1] if uaf_tag in self.UAF_TAG_DICT else uaf_name

        # Print @UAF|@HPU Module EFI Info
        uaf_mod.struct_print(filename=uaf_name + uaf_fext, description=uaf_fdesc, padding=padding + 4)

        # Check if unknown @UAF|@HPU Module Tag is present in @NAL but not in built-in dictionary
        if uaf_tag in nal_dict and uaf_tag not in self.UAF_TAG_DICT and \
                not uaf_tag.startswith(('@ROM', '@R0', '@S0', '@DR', '@DS')):

            printer(message=f'Note: Detected new AMI UCP Module {uaf_tag} ({nal_dict[uaf_tag][1]}) in @NAL!',
                    padding=padding + 4)

        # Generate @UAF|@HPU Module File name, depending on whether decompression will be required
        uaf_sname: str = safe_name(in_name=uaf_name + ('.temp' if is_comp else uaf_fext))

        if uaf_tag in nal_dict:
            uaf_npath: str = safe_path(base_path=extract_path, user_paths=nal_dict[uaf_tag][0])

            make_dirs(in_path=uaf_npath)

            uaf_fname: str = safe_path(base_path=uaf_npath, user_paths=uaf_sname)
        else:
            uaf_fname = safe_path(base_path=extract_path, user_paths=uaf_sname)

        if self.checksum:
            self._chk16_validate(data=uaf_data_all, tag=uaf_tag, padding=padding + 4)

        # Parse Utility Identification Information @UAF|@HPU Module (@UII)
        if uaf_tag == '@UII':
            # Parse @UII Module Raw Structure
            info_hdr: Any = ctypes_struct(buffer=uaf_data_raw, start_offset=0, class_object=UiiHeader)

            # @UII Module Info Data
            info_data: bytes = uaf_data_raw[max(self.UII_HDR_LEN, info_hdr.InfoSize):info_hdr.UIISize]

            # Get @UII Module Info/Description text field
            info_desc: str = info_data.decode('utf-8', 'ignore').strip('\x00 ')

            printer(message='Utility Identification Information:\n', padding=padding + 4)

            info_hdr.struct_print(description=info_desc, padding=padding + 8)  # Print @UII Module Info

            if self.checksum:
                self._chk16_validate(data=uaf_data_raw, tag='@UII > Info', padding=padding + 8)

            # Store/Save @UII Module Info in file
            with open(uaf_fname[:-4] + '.txt', 'a', encoding='utf-8') as uii_out:
                with contextlib.redirect_stdout(uii_out):
                    info_hdr.struct_print(description=info_desc, padding=0)  # Store @UII Module Info

        # Adjust @UAF|@HPU Module Raw Data for extraction
        if is_comp:
            # Some Compressed @UAF|@HPU Module EFI data lack necessary EOF padding
            if uaf_mod.CompressSize > len(uaf_data_raw):
                comp_padd: bytes = b'\x00' * (uaf_mod.CompressSize - len(uaf_data_raw))

                # Add missing padding for decompression
                uaf_data_raw = uaf_data_mod[:self.UAF_MOD_LEN] + uaf_data_raw + comp_padd
            else:
                # Add the EFI/Tiano Compression info before Raw Data
                uaf_data_raw = uaf_data_mod[:self.UAF_MOD_LEN] + uaf_data_raw
        else:
            # No compression, extend to end of Original @UAF|@HPU Module size
            uaf_data_raw = uaf_data_raw[:uaf_mod.OriginalSize]

        # Store/Save @UAF|@HPU Module file
        if uaf_tag != '@UII':  # Skip @UII binary, already parsed
            with open(uaf_fname, 'wb') as uaf_out:
                uaf_out.write(uaf_data_raw)

        # @UAF|@HPU Module EFI/Tiano Decompression
        if is_comp and is_efi_compressed(in_object=uaf_data_raw, strict=False):
            # Decompressed @UAF|@HPU Module file path
            dec_fname: str = uaf_fname.replace('.temp', uaf_fext)

            if efi_decompress(in_path=uaf_fname, out_path=dec_fname, padding=padding + 4):
                with open(dec_fname, 'rb') as dec:
                    uaf_data_raw = dec.read()  # Read back the @UAF|@HPU Module decompressed Raw data

                delete_file(in_path=uaf_fname)  # Successful decompression, delete compressed @UAF|@HPU Module file

                uaf_fname = dec_fname  # Adjust @UAF|@HPU Module file path to the decompressed one

        # Process and Print known text only @UAF|@HPU Modules (after EFI/Tiano Decompression)
        if uaf_tag in self.UAF_TAG_DICT and self.UAF_TAG_DICT[uaf_tag][2] == 'Text':
            printer(message=f'{self.UAF_TAG_DICT[uaf_tag][1]}:', padding=padding + 4)

            printer(message=uaf_data_raw.decode('utf-8', 'ignore'), padding=padding + 8)

        # Parse Default Command Status @UAF|@HPU Module (@DIS)
        if len(uaf_data_raw) and uaf_tag == '@DIS':
            # Parse @DIS Module Raw Header Structure
            dis_hdr: Any = ctypes_struct(buffer=uaf_data_raw, start_offset=0, class_object=DisHeader)

            printer(message='Default Command Status Header:\n', padding=padding + 4)

            dis_hdr.struct_print(padding=padding + 8)  # Print @DIS Module Raw Header Info

            # Store/Save @DIS Module Header Info in file
            with open(uaf_fname[:-3] + 'txt', 'a', encoding='utf-8') as dis:
                with contextlib.redirect_stdout(dis):
                    dis_hdr.struct_print(padding=0)  # Store @DIS Module Header Info

            dis_data: bytes = uaf_data_raw[self.DIS_HDR_LEN:]  # @DIS Module Entries Data

            # Parse all @DIS Module Entries
            for mod_idx in range(dis_hdr.EntryCount):
                # Parse @DIS Module Raw Entry Structure
                dis_mod: Any = ctypes_struct(buffer=dis_data, start_offset=mod_idx * self.DIS_MOD_LEN,
                                             class_object=DisModule)

                printer(message=f'Default Command Status Entry {mod_idx + 1:02d}/{dis_hdr.EntryCount:02d}:\n',
                        padding=padding + 8)

                dis_mod.struct_print(padding=padding + 12)  # Print @DIS Module Raw Entry Info

                # Store/Save @DIS Module Entry Info in file
                with open(uaf_fname[:-3] + 'txt', 'a', encoding='utf-8') as dis:
                    with contextlib.redirect_stdout(dis):
                        printer(message=None)

                        dis_mod.struct_print(padding=4)  # Store @DIS Module Entry Info

            delete_file(in_path=uaf_fname)  # Delete @DIS Module binary, info exported as text

        # Parse Name List @UAF|@HPU Module (@NAL)
        if len(uaf_data_raw) >= 5 and (uaf_tag, uaf_data_raw[0], uaf_data_raw[4]) == ('@NAL', 0x40, 0x3A):
            nal_info: list[str] = uaf_data_raw.decode('utf-8',
                                                      errors='ignore').replace('\r', '').strip().split('\n')

            printer(message='AMI UCP Module Name List:\n', padding=padding + 4)

            # Parse all @NAL Module Entries
            for info in nal_info:
                info_tag, info_value = info.split(':', 1)

                # Print @NAL Module Tag-Path Info
                printer(message=f'{info_tag} : {info_value}', padding=padding + 8, new_line=False)

                # Split OS-agnostic path in parts
                info_part: Any = agnostic_path(in_path=info_value).parts

                # Get path without drive/root or file
                info_path: str = to_string(in_object=info_part[1:-1], sep_char=os.sep)

                # Get file from last path part
                info_name: str = info_part[-1]

                # Assign a file path & name to each Tag
                nal_dict[info_tag] = (info_path, info_name)

        # Parse Insyde BIOS @UAF|@HPU Module (@INS)
        if uaf_tag == '@INS':
            ins_dir: str = os.path.join(extract_path, safe_name(in_name=f'{uaf_tag}_nested-IFD'))

            insyde_ifd_extract: InsydeIfdExtract = InsydeIfdExtract(
                input_object=uaf_fname, extract_path=extract_folder(ins_dir), padding=padding + 4)

            if insyde_ifd_extract.check_format():
                if insyde_ifd_extract.parse_format():
                    delete_file(in_path=uaf_fname)  # Delete raw nested Insyde IFD image after successful extraction

        pfat_dir: str = os.path.join(extract_path, safe_name(in_name=uaf_name))

        ami_pfat_extract: AmiPfatExtract = AmiPfatExtract(
            input_object=uaf_data_raw, extract_path=extract_folder(pfat_dir), padding=padding + 4)

        # Detect & Unpack AMI BIOS Guard (PFAT) BIOS image
        if ami_pfat_extract.check_format():
            ami_pfat_extract.parse_format()

            delete_file(in_path=uaf_fname)  # Delete raw PFAT BIOS image after successful extraction

        # Detect Intel Engine firmware image and show ME Analyzer advice
        if uaf_tag.startswith('@ME') and PAT_INTEL_ENGINE.search(uaf_data_raw):
            printer(message='Intel Management Engine (ME) Firmware:\n', padding=padding + 4)
            printer(message='Use "ME Analyzer" from https://github.com/platomav/MEAnalyzer',
                    padding=padding + 8, new_line=False)

        uaf_dir: str = extract_folder(os.path.join(extract_path, safe_name(in_name=f'{uaf_tag}_nested-UCP')))

        ami_ucp_extract: AmiUcpExtract = AmiUcpExtract(
            input_object=uaf_data_raw, extract_path=uaf_dir, padding=padding + 4, checksum=self.checksum)

        # Parse Nested AMI UCP image
        if ami_ucp_extract.check_format():
            ami_ucp_extract.parse_format()

            delete_file(in_path=uaf_fname)  # Delete raw nested AMI UCP image after successful extraction

        return nal_dict



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Copyright (C) 2022-2024 Plato Mavropoulos
"""


def to_string(in_object: str | list | tuple, sep_char: str = '') -> str:
    """ Get string from given input object """

    if isinstance(in_object, (list, tuple)):
        out_string: str = sep_char.join(map(str, in_object))
    else:
        out_string = str(in_object)

    return out_string


def to_ordinal(in_number: int) -> str:
    """
    Get ordinal (textual) representation of input numerical value

    https://leancrew.com/all-this/2020/06/ordinals-in-python/ by Dr. Drang
    """

    ordinals: list[str] = ['th', 'st', 'nd', 'rd'] + ['th'] * 10

    numerical: int = in_number % 100

    if numerical > 13:
        return f'{in_number}{ordinals[numerical % 10]}'

    return f'{in_number}{ordinals[numerical]}'


def file_to_bytes(in_object: str | bytes | bytearray) -> bytes:
    """ Get bytes from given buffer or file path """

    if isinstance(in_object, str):
        with open(in_object, 'rb') as object_fp:
            return object_fp.read()

    return bytes(in_object)


def bytes_to_hex(in_buffer: bytes, order: str, data_len: int, slice_len: int | None = None) -> str:
    """ Converts bytes to hex string, controlling endianess, data size and string slicing """

    # noinspection PyTypeChecker
    return f'{int.from_bytes(in_buffer, byteorder=order):0{data_len * 2}X}'[:slice_len]  # type: ignore


def remove_quotes(in_text: str) -> str:
    """ Remove leading/trailing quotes from path """

    out_text: str = to_string(in_object=in_text).strip()

    if len(out_text) >= 2:
        if (out_text[0] == '"' and out_text[-1] == '"') or (out_text[0] == "'" and out_text[-1] == "'"):
            out_text = out_text[1:-1]

    return out_text


def to_boxed(in_text: str) -> str:
    """ Box string into two horizontal lines of same size """

    box_line: str = '-' * len(to_string(in_object=in_text))

    return f'{box_line}\n{in_text}\n{box_line}'



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Panasonic BIOS Extract
Panasonic BIOS Package Extractor
Copyright (C) 2018-2024 Plato Mavropoulos
"""

import io
import logging
import os
import re

from typing import Final

import pefile

# noinspection PyPackageRequirements
from dissect.util.compression import lznt1

from biosutilities.common.compression import is_szip_supported, szip_decompress
from biosutilities.common.paths import delete_file, is_file_read, path_files, make_dirs, path_stem, safe_name
from biosutilities.common.executables import ms_pe_desc, ms_pe, is_ms_pe, ms_pe_info_show
from biosutilities.common.patterns import PAT_MICROSOFT_CAB
from biosutilities.common.system import printer
from biosutilities.common.templates import BIOSUtility
from biosutilities.common.texts import file_to_bytes

from biosutilities.ami_pfat_extract import AmiPfatExtract


class PanasonicBiosExtract(BIOSUtility):
    """ Panasonic BIOS Package Extractor """

    TITLE: str = 'Panasonic BIOS Package Extractor'

    PAN_PE_DESC_UNP: Final[str] = 'UNPACK UTILITY'

    PAN_PE_DESC_UPD: Final[str] = 'BIOS UPDATE'

    def check_format(self) -> bool:
        """ Check if input is Panasonic BIOS Package PE """

        pe_file: pefile.PE | None = ms_pe(in_file=self.input_object, silent=True)

        if not pe_file:
            return False

        if ms_pe_desc(pe_file=pe_file, silent=True).decode('utf-8', 'ignore').upper() not in (
                self.PAN_PE_DESC_UNP, self.PAN_PE_DESC_UPD):
            return False

        return True

    def parse_format(self) -> bool:
        """ Parse & Extract Panasonic BIOS Package PE """

        upd_pe_file: pefile.PE = ms_pe(in_file=self.input_object, padding=self.padding)  # type: ignore

        upd_pe_name: str = self._panasonic_pkg_name(input_object=self.input_object)

        printer(message=f'Panasonic BIOS Package > PE ({upd_pe_name})\n'.replace(' ()', ''), padding=self.padding)

        ms_pe_info_show(pe_file=upd_pe_file, padding=self.padding + 4)

        make_dirs(in_path=self.extract_path)

        upd_pe_path: str = self._panasonic_cab_extract(input_object=self.input_object,
                                                       extract_path=self.extract_path, padding=self.padding + 8)

        upd_padding: int = self.padding

        if upd_pe_path:
            upd_padding = self.padding + 16

            upd_pe_name = self._panasonic_pkg_name(input_object=upd_pe_path)

            printer(message=f'Panasonic BIOS Update > PE ({upd_pe_name})\n'.replace(' ()', ''), padding=upd_padding)

            upd_pe_file = ms_pe(in_file=upd_pe_path, padding=upd_padding)  # type: ignore

            ms_pe_info_show(pe_file=upd_pe_file, padding=upd_padding + 4)

            delete_file(in_path=upd_pe_path)

        is_upd_extracted: bool = self._panasonic_res_extract(pe_file=upd_pe_file, extract_path=self.extract_path,
                                                             pe_name=upd_pe_name, padding=upd_padding + 8)

        if not is_upd_extracted:
            is_upd_extracted = self._panasonic_img_extract(pe_file=upd_pe_file, extract_path=self.extract_path,
                                                           pe_name=upd_pe_name, padding=upd_padding + 8)

        return is_upd_extracted

    @staticmethod
    def _panasonic_pkg_name(input_object: str | bytes | bytearray) -> str:
        """ Get Panasonic BIOS Package file name, when applicable """

        if isinstance(input_object, str) and is_file_read(in_path=input_object):
            return safe_name(in_name=path_stem(in_path=input_object))

        return ''

    def _panasonic_cab_extract(self, input_object: str | bytes | bytearray, extract_path: str, padding: int = 0) -> str:
        """ Search and Extract Panasonic BIOS Package PE CAB archive """

        input_data: bytes = file_to_bytes(in_object=input_object)

        cab_match: re.Match[bytes] | None = PAT_MICROSOFT_CAB.search(input_data)

        if cab_match:
            cab_bgn: int = cab_match.start()

            cab_end: int = cab_bgn + int.from_bytes(input_data[cab_bgn + 0x8:cab_bgn + 0xC], byteorder='little')

            cab_tag: str = f'[0x{cab_bgn:06X}-0x{cab_end:06X}]'

            cab_path: str = os.path.join(extract_path, f'CAB_{cab_tag}.cab')

            with open(cab_path, 'wb') as cab_file_object:
                cab_file_object.write(input_data[cab_bgn:cab_end])

            if is_szip_supported(in_path=cab_path):
                printer(message=f'Panasonic BIOS Package > PE > CAB {cab_tag}', padding=padding)

                if szip_decompress(in_path=cab_path, out_path=extract_path, in_name='CAB',
                                   padding=padding + 4, check=True):
                    delete_file(in_path=cab_path)  # Successful extraction, delete CAB archive

                    for extracted_file_path in path_files(in_path=extract_path):
                        if is_file_read(in_path=extracted_file_path):
                            extracted_pe_file: pefile.PE | None = ms_pe(
                                in_file=extracted_file_path, padding=padding, silent=True)

                            if extracted_pe_file:
                                extracted_pe_desc: bytes = ms_pe_desc(pe_file=extracted_pe_file, silent=True)

                                if extracted_pe_desc.decode('utf-8', 'ignore'
                                                            ).upper() == self.PAN_PE_DESC_UPD:
                                    return extracted_file_path

        return ''

    @staticmethod
    def _panasonic_res_extract(pe_file: pefile.PE, extract_path: str, pe_name: str = '', padding: int = 0) -> bool:
        """ Extract & Decompress Panasonic BIOS Update PE RCDATA (LZNT1) """

        is_rcdata: bool = False

        # When fast_load is used, IMAGE_DIRECTORY_ENTRY_RESOURCE must be parsed prior to RCDATA Directories
        pe_file.parse_data_directories(directories=[pefile.DIRECTORY_ENTRY['IMAGE_DIRECTORY_ENTRY_RESOURCE']])

        # noinspection PyUnresolvedReferences
        for entry in pe_file.DIRECTORY_ENTRY_RESOURCE.entries:
            # Parse all Resource Data Directories > RCDATA (ID = 10)
            if entry.struct.name == 'IMAGE_RESOURCE_DIRECTORY_ENTRY' and entry.struct.Id == 0xA:
                is_rcdata = True

                for resource in entry.directory.entries:
                    res_bgn: int = resource.directory.entries[0].data.struct.OffsetToData
                    res_len: int = resource.directory.entries[0].data.struct.Size
                    res_end: int = res_bgn + res_len

                    res_bin: bytes = pe_file.get_data(res_bgn, res_len)

                    res_tag: str = f'{pe_name} [0x{res_bgn:06X}-0x{res_end:06X}]'.strip()

                    res_out: str = os.path.join(extract_path, f'{res_tag}')

                    printer(message=res_tag, padding=padding)

                    try:
                        res_raw: bytes = lznt1.decompress(src=res_bin[0x8:])

                        if len(res_raw) != int.from_bytes(res_bin[0x4:0x8], byteorder='little'):
                            raise ValueError('LZNT1_DECOMPRESS_BAD_SIZE')

                        printer(message='Successful LZNT1 decompression via Dissect!', padding=padding + 4)
                    except Exception as error:  # pylint: disable=broad-except
                        logging.debug('Error: LZNT1 decompression of %s failed: %s', res_tag, error)

                        res_raw = res_bin

                        printer(message='Successful PE Resource extraction!', padding=padding + 4)

                    pfat_dir: str = os.path.join(extract_path, res_tag)

                    ami_pfat_extract: AmiPfatExtract = AmiPfatExtract(
                        input_object=res_raw, extract_path=pfat_dir, padding=padding + 8)

                    # Detect & Unpack AMI BIOS Guard (PFAT) BIOS image
                    if ami_pfat_extract.check_format():
                        ami_pfat_extract.parse_format()
                    else:
                        if is_ms_pe(in_file=res_raw):
                            res_ext: str = 'exe'
                        elif res_raw.startswith(b'[') and res_raw.endswith((b'\x0D\x0A', b'\x0A')):
                            res_ext = 'txt'
                        else:
                            res_ext = 'bin'

                        if res_ext == 'txt':
                            printer(message=None, new_line=False)

                            for line in io.BytesIO(res_raw).readlines():
                                line_text: str = line.decode('utf-8', 'ignore').rstrip()

                                printer(message=line_text, padding=padding + 8, new_line=False)

                        with open(f'{res_out}.{res_ext}', 'wb') as out_file_object:
                            out_file_object.write(res_raw)

        return is_rcdata

    @staticmethod
    def _panasonic_img_extract(pe_file: pefile.PE, extract_path: str, pe_name: str = '',
                               padding: int = 0) -> bool:
        """ Extract Panasonic BIOS Update PE Data when RCDATA is not available """

        pe_data: bytes = bytes(pe_file.__data__)

        sec_bgn: int = pe_file.OPTIONAL_HEADER.DATA_DIRECTORY[pefile.DIRECTORY_ENTRY[
            'IMAGE_DIRECTORY_ENTRY_SECURITY']].VirtualAddress

        img_bgn: int = (pe_file.OPTIONAL_HEADER.BaseOfData +  # type: ignore
                        pe_file.OPTIONAL_HEADER.SizeOfInitializedData)

        img_end: int = sec_bgn or len(pe_data)

        img_bin: bytes = pe_data[img_bgn:img_end]

        img_tag: str = f'{pe_name} [0x{img_bgn:X}-0x{img_end:X}]'.strip()

        img_out: str = os.path.join(extract_path, f'{img_tag}.bin')

        printer(message=img_tag, padding=padding)

        with open(img_out, 'wb') as out_img_object:
            out_img_object.write(img_bin)

        printer(message='Successful PE Data extraction!', padding=padding + 4)

        return bool(img_bin)



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Apple EFI ID
Apple EFI Image Identifier
Copyright (C) 2018-2024 Plato Mavropoulos
"""

import ctypes
import os
import struct
import subprocess
import zlib

from datetime import datetime
from re import Match
from typing import Any, Final

from biosutilities.common.externals import uefiextract_path, uefifind_path
from biosutilities.common.paths import (delete_dirs, delete_file, is_file_read, make_dirs, path_size,
                                        path_suffixes, runtime_root)
from biosutilities.common.patterns import PAT_INTEL_IBIOSI, PAT_APPLE_ROM_VER
from biosutilities.common.structs import CHAR, ctypes_struct, UINT8
from biosutilities.common.system import printer
from biosutilities.common.templates import BIOSUtility

EFI_EXTENSIONS: Final[list[str]] = ['.fd', '.scap', '.im4p']
EFI_MAX_SIZE: Final[int] = 0x3000000


class IntelBiosId(ctypes.LittleEndianStructure):
    """
    Intel BIOS ID Structure

    https://github.com/tianocore/edk2-platforms/blob/master/Platform/Intel/BoardModulePkg/Include/Guid/BiosId.h
    """

    _pack_ = 1
    _fields_ = [
        ('Signature',       CHAR * 8),      # 0x00
        ('BoardID',         UINT8 * 16),    # 0x08
        ('Dot1',            UINT8 * 2),     # 0x18
        ('BoardExt',        UINT8 * 6),     # 0x1A
        ('Dot2',            UINT8 * 2),     # 0x20
        ('VersionMajor',    UINT8 * 8),     # 0x22
        ('Dot3',            UINT8 * 2),     # 0x2A
        ('BuildType',       UINT8 * 2),     # 0x2C
        ('VersionMinor',    UINT8 * 4),     # 0x2E
        ('Dot4',            UINT8 * 2),     # 0x32
        ('Year',            UINT8 * 4),     # 0x34
        ('Month',           UINT8 * 4),     # 0x38
        ('Day',             UINT8 * 4),     # 0x3C
        ('Hour',            UINT8 * 4),     # 0x40
        ('Minute',          UINT8 * 4),     # 0x44
        ('NullTerminator',  UINT8 * 2)      # 0x48
        # 0x4A
    ]

    @staticmethod
    def _decode(field: bytes) -> str:
        return struct.pack('B' * len(field), *field).decode('utf-16', 'ignore').strip('\x00 ')

    def get_bios_id(self) -> dict[str, str]:
        """ Get Apple/Intel EFI BIOS ID """

        board_id: str = self._decode(field=self.BoardID)
        board_ext: str = self._decode(field=self.BoardExt)
        version_major: str = self._decode(field=self.VersionMajor)
        build_type: str = self._decode(field=self.BuildType)
        version_minor: str = self._decode(field=self.VersionMinor)
        build_year: int = int(f'20{self._decode(field=self.Year)}')
        build_month: int = int(self._decode(field=self.Month))
        build_day: int = int(self._decode(field=self.Day))
        build_hour: int = int(self._decode(field=self.Hour))
        build_minute: int = int(self._decode(field=self.Minute))

        efi_version: str = f'{version_major}_{build_type}{version_minor}'
        efi_date: datetime = datetime(build_year, build_month, build_day, build_hour, build_minute)
        efi_name: str = f'{board_id}_{board_ext}_{efi_version}_{efi_date.strftime("%Y-%m-%d_%H-%M")}'

        return {
            'board_id': board_id,
            'apple_id': board_ext,
            'version': efi_version,
            'date': efi_date.isoformat(),
            'name': efi_name
        }

    def struct_print(self, padding: int = 0) -> None:
        """ Display structure information """

        ibiosi: dict[str, str] = self.get_bios_id()

        printer(message=['Board ID: ', ibiosi['board_id']], padding=padding, new_line=False)
        printer(message=['Apple ID: ', ibiosi['apple_id']], padding=padding, new_line=False)
        printer(message=['Version:  ', ibiosi['version']], padding=padding, new_line=False)
        printer(message=['Datetime: ', ibiosi['date']], padding=padding, new_line=False)


class AppleEfiIdentify(BIOSUtility):
    """ Apple EFI Image Identifier """

    TITLE: str = 'Apple EFI Image Identifier'

    PAT_UEFIFIND: Final[str] = f'244942494F534924{"." * 32}2E00{"." * 12}2E00{"." * 16}2E00{"." * 12}2E00{"." * 40}00'

    def __init__(self, input_object: str | bytes | bytearray = b'', extract_path: str = '', padding: int = 0,
                 silent: bool = False) -> None:
        super().__init__(input_object=input_object, extract_path=extract_path, padding=padding)

        self.silent: bool = silent

        self.efi_file_name: str = ''

        self.intel_bios_info: dict[str, str] = {
            'board_id': '',
            'apple_id': '',
            'version': '',
            'date': '',
            'name': ''
        }

        self.apple_rom_version: dict[str, str] = {
            'bios_id': '',
            'board_id': '',
            'build_type': '',
            'buildcave_id': '',
            'built_by': '',
            'compiler': '',
            'date': '',
            'efi_version': '',
            'model': '',
            'rom_version': '',
            'revision': '',
            'uuid': ''
        }

    def check_format(self) -> bool:
        """ Check if input is Apple EFI image """

        if isinstance(self.input_object, str) and is_file_read(in_path=self.input_object):
            if path_suffixes(in_path=self.input_object)[-1].lower() not in EFI_EXTENSIONS:
                return False

            if path_size(in_path=self.input_object) > EFI_MAX_SIZE:
                return False

            input_path: str = self.input_object
        else:
            if len(self.input_buffer) > EFI_MAX_SIZE:
                return False

            input_path = os.path.join(runtime_root(), 'APPLE_EFI_ID_INPUT_BUFFER_CHECK.tmp')

            with open(input_path, 'wb') as check_out:
                check_out.write(self.input_buffer)

        if PAT_INTEL_IBIOSI.search(self.input_buffer):
            return True

        uefifind_cmd: list[str] = [uefifind_path(), input_path, 'body', 'list', self.PAT_UEFIFIND]

        uefifind_res: subprocess.CompletedProcess[bytes] = subprocess.run(
            uefifind_cmd, check=False, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

        if input_path != self.input_object:
            delete_file(in_path=input_path)

        if uefifind_res.returncode == 0:
            return True

        return False

    def parse_format(self) -> bool:
        """ Parse & Identify (or Rename) Apple EFI image """

        if isinstance(self.input_object, str) and is_file_read(in_path=self.input_object):
            input_path: str = self.input_object
        else:
            input_path = os.path.join(runtime_root(), 'APPLE_EFI_ID_INPUT_BUFFER_PARSE.tmp')

            with open(input_path, 'wb') as parse_out:
                parse_out.write(self.input_buffer)

        bios_id_match: Match[bytes] | None = PAT_INTEL_IBIOSI.search(self.input_buffer)

        if bios_id_match:
            bios_id_res: str = f'0x{bios_id_match.start():X}'

            bios_id_hdr: Any = ctypes_struct(buffer=self.input_buffer, start_offset=bios_id_match.start(),
                                             class_object=IntelBiosId)
        else:
            # The $IBIOSI$ pattern is within EFI compressed modules so we need to use UEFIFind and UEFIExtract

            bios_id_res = subprocess.check_output([uefifind_path(), input_path, 'body', 'list',
                                                   self.PAT_UEFIFIND], text=True)[:36]

            make_dirs(in_path=self.extract_path)

            uefiextract_dir: str = os.path.join(self.extract_path, 'uefiextract_temp')

            # UEFIExtract must create its output folder itself
            delete_dirs(in_path=uefiextract_dir)

            _ = subprocess.run([uefiextract_path(), input_path, bios_id_res, '-o', uefiextract_dir, '-m', 'body'],
                               check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

            with open(os.path.join(uefiextract_dir, 'body.bin'), 'rb') as raw_body:
                body_buffer: bytes = raw_body.read()

            # Detect decompressed $IBIOSI$ pattern
            bios_id_match = PAT_INTEL_IBIOSI.search(body_buffer)

            if not bios_id_match:
                raise RuntimeError('Failed to detect decompressed $IBIOSI$ pattern!')

            bios_id_hdr = ctypes_struct(buffer=body_buffer, start_offset=bios_id_match.start(),
                                        class_object=IntelBiosId)

            delete_dirs(in_path=uefiextract_dir)  # Successful UEFIExtract extraction, remove its output folder

        if not self.silent:
            printer(message=f'Detected Intel BIOS Info at {bios_id_res}\n', padding=self.padding)

            bios_id_hdr.struct_print(padding=self.padding + 4)

        self.intel_bios_info |= bios_id_hdr.get_bios_id()

        self.efi_file_name = (f'{self.intel_bios_info["name"]}_{zlib.adler32(self.input_buffer):08X}'
                              f'{path_suffixes(in_path=input_path)[-1]}')

        _ = self._apple_rom_version(input_buffer=self.input_buffer, padding=self.padding)

        if input_path != self.input_object:
            delete_file(in_path=input_path)

        return True

    def _apple_rom_version(self, input_buffer: bytes | bytearray, padding: int = 0) -> bool:
        rom_version_match: Match[bytes] | None = PAT_APPLE_ROM_VER.search(input_buffer)

        if rom_version_match:
            rom_version_match_off: int = rom_version_match.start()

            rom_version_header_len: int = input_buffer[rom_version_match_off:].find(b'\n')

            if rom_version_header_len != -1:
                rom_version_data_bgn: int = rom_version_match_off + rom_version_header_len

                rom_version_data_len: int = input_buffer[rom_version_data_bgn:].find(b'\x00')

                if rom_version_data_len != -1:
                    rom_version_data_end: int = rom_version_data_bgn + rom_version_data_len

                    rom_version_data: bytes = input_buffer[rom_version_data_bgn:rom_version_data_end]

                    rom_version_text: str = rom_version_data.decode('utf-8').strip('\n')

                    if not self.silent:
                        printer(message=f'Detected Apple ROM Version at 0x{rom_version_match_off:X}', padding=padding)

                        printer(message=rom_version_text, strip=True, padding=padding + 4)

                    for rom_version_line in [line.strip() for line in rom_version_text.split('\n')]:
                        rom_version_parts: list[str] = rom_version_line.split(sep=':', maxsplit=1)

                        rom_version_key: str = rom_version_parts[0].strip().lower().replace(' ', '_')
                        rom_version_val: str = rom_version_parts[1].strip()

                        if rom_version_key == 'uuid':
                            if self.apple_rom_version[rom_version_key] == '':
                                self.apple_rom_version[rom_version_key] = rom_version_val
                            else:
                                self.apple_rom_version[rom_version_key] += f', {rom_version_val}'
                        elif rom_version_key == 'date':
                            self.apple_rom_version[rom_version_key] = datetime.strptime(rom_version_val.replace(
                                ' PDT', '').replace(' PST', ''), '%a %b %d %H:%M:%S %Y').isoformat()
                        else:
                            self.apple_rom_version[rom_version_key] = rom_version_val

                    return True

        return False



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Apple EFI PKG
Apple EFI Package Extractor
Copyright (C) 2019-2024 Plato Mavropoulos
"""

import os

from biosutilities.common.compression import is_szip_supported, szip_decompress
from biosutilities.common.paths import (copy_file, delete_dirs, delete_file, extract_folder, is_dir,
                                        is_file_read, make_dirs, path_files, path_name, path_stem,
                                        path_suffixes, runtime_root)
from biosutilities.common.system import printer
from biosutilities.common.templates import BIOSUtility

from biosutilities.apple_efi_id import AppleEfiIdentify, EFI_EXTENSIONS
from biosutilities.apple_efi_im4p import AppleEfiIm4pSplit
from biosutilities.apple_efi_pbzx import AppleEfiPbzxExtract


class AppleEfiPkgExtract(BIOSUtility):
    """ Apple EFI Package Extractor """

    TITLE: str = 'Apple EFI Package Extractor'

    def check_format(self) -> bool:
        """ Check if input is Apple EFI PKG package """

        is_apple_efi_pkg: bool = False

        if isinstance(self.input_object, str) and self._is_file_processable(input_path=self.input_object):
            input_path: str = self.input_object
        else:
            input_path = os.path.join(runtime_root(), 'APPLE_EFI_PKG_INPUT_BUFFER_CHECK.bin')

            with open(input_path, 'wb') as input_path_object:
                input_path_object.write(self.input_buffer)

        for pkg_type in ('XAR', 'TAR', 'DMG'):
            if is_szip_supported(in_path=input_path, args=[f'-t{pkg_type}:s0']):
                is_apple_efi_pkg = True

                break

        if input_path != self.input_object:
            delete_file(in_path=input_path)

        return is_apple_efi_pkg

    def parse_format(self) -> bool:
        """ Parse & Extract Apple EFI PKG packages """

        if isinstance(self.input_object, str) and self._is_file_processable(input_path=self.input_object):
            input_path: str = self.input_object
        else:
            input_path = os.path.join(runtime_root(), 'APPLE_EFI_PKG_INPUT_BUFFER_PARSE.bin')

            with open(input_path, 'wb') as input_path_object:
                input_path_object.write(self.input_buffer)

        working_dir: str = os.path.join(self.extract_path, 'temp')

        make_dirs(in_path=working_dir)

        for pkg_type in ('XAR', 'TAR', 'DMG'):
            if is_szip_supported(in_path=input_path, args=[f'-t{pkg_type}']):
                if szip_decompress(in_path=input_path, out_path=working_dir, in_name=pkg_type, padding=self.padding,
                                   args=None if pkg_type == 'DMG' else [f'-t{pkg_type}']):
                    break
        else:
            return False

        if input_path != self.input_object:
            delete_file(in_path=input_path)

        for work_file in path_files(in_path=working_dir):
            if self._is_file_processable(input_path=work_file):
                self._pbzx_zip(input_path=work_file, padding=self.padding + 4)
                self._gzip_cpio(input_path=work_file, padding=self.padding + 4)
                self._dmg_zip(input_path=work_file, padding=self.padding + 4)
                self._xar_gzip(input_path=work_file, padding=self.padding + 4)

        delete_dirs(in_path=working_dir)

        return True

    @staticmethod
    def _is_file_processable(input_path: str) -> bool:
        return is_file_read(in_path=input_path) and path_stem(in_path=input_path) != '.aea'

    def _xar_gzip(self, input_path: str, padding: int = 0) -> None:
        """ XAR/TAR > GZIP """

        for pkg_type in ('XAR', 'TAR'):
            if is_szip_supported(in_path=input_path, args=[f'-t{pkg_type}']):
                pkg_path: str = extract_folder(in_path=input_path, suffix=f'_{pkg_type.lower()}_gzip')

                if szip_decompress(in_path=input_path, out_path=pkg_path, in_name=pkg_type,
                                   padding=padding, args=[f'-t{pkg_type}']):
                    for pkg_file in path_files(in_path=pkg_path):
                        if self._is_file_processable(input_path=pkg_file):
                            self._gzip_cpio(input_path=pkg_file, padding=padding + 4)

                break

    def _dmg_zip(self, input_path: str, padding: int = 0) -> None:
        """ DMG > ZIP """

        if is_szip_supported(in_path=input_path, args=['-tDMG']):
            dmg_path: str = extract_folder(in_path=input_path, suffix='_dmg_zip')

            if szip_decompress(in_path=input_path, out_path=dmg_path, in_name='DMG', padding=padding, args=None):
                for dmg_file in path_files(in_path=dmg_path):
                    if self._is_file_processable(input_path=dmg_file):
                        if is_szip_supported(in_path=dmg_file, args=['-tZIP']):
                            zip_path: str = extract_folder(in_path=dmg_file)

                            if szip_decompress(in_path=dmg_file, out_path=zip_path, in_name='ZIP',
                                               padding=padding + 4, args=['-tZIP']):
                                for zip_file in path_files(in_path=zip_path):
                                    if self._is_file_processable(input_path=zip_file):
                                        self._im4p_id(input_path=zip_file, padding=padding + 8)

    def _pbzx_zip(self, input_path: str, padding: int = 0) -> None:
        """ PBZX > ZIP """

        pbzx_path: str = extract_folder(in_path=input_path, suffix='_pbzx_zip')

        pbzx_module: AppleEfiPbzxExtract = AppleEfiPbzxExtract(
            input_object=input_path, extract_path=pbzx_path, padding=padding + 4)

        if pbzx_module.check_format():
            printer(message=f'Extracting PBZX via {pbzx_module.TITLE}', padding=padding)

            if pbzx_module.parse_format():
                printer(message=f'Successful PBZX extraction via {pbzx_module.TITLE}!', padding=padding)

                for pbzx_file in path_files(in_path=pbzx_path):
                    if self._is_file_processable(input_path=pbzx_file):
                        if is_szip_supported(in_path=pbzx_file, args=['-tZIP']):
                            zip_path: str = extract_folder(in_path=pbzx_file)

                            if szip_decompress(in_path=pbzx_file, out_path=zip_path, in_name='ZIP',
                                               padding=padding + 4, args=['-tZIP']):
                                for zip_file in path_files(in_path=zip_path):
                                    if self._is_file_processable(input_path=zip_file):
                                        self._im4p_id(input_path=zip_file, padding=padding + 8)

    def _gzip_cpio(self, input_path: str, padding: int = 0) -> None:
        """ GZIP > CPIO """

        if is_szip_supported(in_path=input_path, args=['-tGZIP']):
            gzip_path: str = extract_folder(in_path=input_path, suffix='_gzip_cpio')

            if szip_decompress(in_path=input_path, out_path=gzip_path, in_name='GZIP',
                               padding=padding, args=['-tGZIP']):
                for gzip_file in path_files(in_path=gzip_path):
                    if self._is_file_processable(input_path=gzip_file):
                        if is_szip_supported(in_path=gzip_file, args=['-tCPIO']):
                            cpio_path: str = extract_folder(in_path=gzip_file)

                            if szip_decompress(in_path=gzip_file, out_path=cpio_path, in_name='CPIO',
                                               padding=padding + 4, args=['-tCPIO']):
                                for cpio_file in path_files(in_path=cpio_path):
                                    if self._is_file_processable(input_path=cpio_file):
                                        self._im4p_id(input_path=cpio_file, padding=padding + 8)

    def _im4p_id(self, input_path: str, padding: int = 0) -> None:
        """ Split IM4P (if applicable), identify and copy EFI """

        if path_suffixes(in_path=input_path)[-1].lower() not in EFI_EXTENSIONS:
            return None

        if not self._is_file_processable(input_path=input_path):
            return None

        working_dir: str = extract_folder(in_path=input_path)

        if not AppleEfiIdentify(input_object=input_path, extract_path=working_dir).check_format():
            return None

        printer(message=path_name(in_path=input_path), padding=padding)

        im4p_module: AppleEfiIm4pSplit = AppleEfiIm4pSplit(
            input_object=input_path, extract_path=working_dir, padding=padding + 8)

        if im4p_module.check_format():
            printer(message=f'Splitting IM4P via {im4p_module.TITLE}', padding=padding + 4)

            im4p_module.parse_format()

        for efi_path in path_files(in_path=working_dir) if is_dir(in_path=working_dir) else [input_path]:
            if self._is_file_processable(input_path=efi_path):
                efi_id_module: AppleEfiIdentify = AppleEfiIdentify(
                    input_object=efi_path, extract_path=extract_folder(in_path=efi_path), padding=padding + 8)

                if efi_id_module.check_format():
                    printer(message=f'Identifying EFI via {efi_id_module.TITLE}', padding=padding + 4)

                    if efi_id_module.parse_format():
                        efi_path_final: str = os.path.join(self.extract_path, efi_id_module.efi_file_name)

                        copy_file(in_path=efi_path, out_path=efi_path_final, metadata=True)

        delete_dirs(in_path=working_dir)

        return None



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Fujitsu UPC Extract
Fujitsu UPC BIOS Extractor
Copyright (C) 2021-2024 Plato Mavropoulos
"""

import os

from biosutilities.common.compression import efi_decompress, is_efi_compressed
from biosutilities.common.paths import delete_file, make_dirs, is_file_read, path_name, path_suffixes
from biosutilities.common.templates import BIOSUtility


class FujitsuUpcExtract(BIOSUtility):
    """ Fujitsu UPC BIOS Extractor """

    TITLE: str = 'Fujitsu UPC BIOS Extractor'

    def check_format(self) -> bool:
        """ Check if input is Fujitsu UPC image """
        is_upc: bool = False

        if isinstance(self.input_object, str) and is_file_read(in_path=self.input_object):
            is_upc = path_suffixes(self.input_object)[-1].upper() == '.UPC'
        elif isinstance(self.input_object, (bytes, bytearray)):
            is_upc = True

        if is_upc:
            is_upc = is_efi_compressed(in_object=self.input_object)

        return is_upc

    def parse_format(self) -> bool:
        """ Parse & Extract Fujitsu UPC image """

        make_dirs(in_path=self.extract_path)

        if isinstance(self.input_object, str) and is_file_read(in_path=self.input_object):
            input_name: str = path_name(in_path=self.input_object)

            input_path: str = self.input_object

            if input_name.upper().endswith('.UPC'):
                input_name = input_name[:-4]
        else:
            input_name = 'Fujitsu_UPC_Image'

            input_path = os.path.join(self.extract_path, f'{input_name}.UPC')

            with open(input_path, 'wb') as input_path_object:
                input_path_object.write(self.input_buffer)

        output_path: str = os.path.join(self.extract_path, f'{input_name}.bin')

        efi_status: bool = efi_decompress(in_path=input_path, out_path=output_path, padding=self.padding)

        if input_path != self.input_object:
            delete_file(in_path=input_path)

        return efi_status



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Copyright (C) 2022-2024 Plato Mavropoulos
"""

from biosutilities.common.paths import runtime_root
from biosutilities.common.texts import file_to_bytes


class BIOSUtility:
    """ Base class for BIOSUtilities """

    TITLE: str = 'BIOS Utility'

    def __init__(self, input_object: str | bytes | bytearray = b'', extract_path: str = runtime_root(),
                 padding: int = 0) -> None:
        self.input_object: str | bytes | bytearray = input_object
        self.extract_path: str = extract_path
        self.padding: int = padding

        self.__input_buffer: bytes = b''

    @property
    def input_buffer(self) -> bytes:
        """ Get input object buffer """

        if not self.__input_buffer:
            self.__input_buffer = file_to_bytes(in_object=self.input_object)

        return self.__input_buffer

    def check_format(self) -> bool:
        """ Check if input object is of specific supported format """

        raise NotImplementedError(f'Method "check_format" not implemented at {__name__}')

    def parse_format(self) -> bool:
        """ Process input object as a specific supported format """

        raise NotImplementedError(f'Method "parse_format" not implemented at {__name__}')



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Copyright (C) 2022-2024 Plato Mavropoulos
"""

import ctypes

from typing import Any, Final

CHAR: Final[type[ctypes.c_char] | int] = ctypes.c_char
UINT8: Final[type[ctypes.c_ubyte] | int] = ctypes.c_ubyte
UINT16: Final[type[ctypes.c_ushort] | int] = ctypes.c_ushort
UINT32: Final[type[ctypes.c_uint] | int] = ctypes.c_uint
UINT64: Final[type[ctypes.c_uint64] | int] = ctypes.c_uint64


def ctypes_struct(buffer: bytes | bytearray, start_offset: int, class_object: Any,
                  param_list: list | None = None) -> Any:
    """
    https://github.com/skochinsky/me-tools/blob/master/me_unpack.py by Igor Skochinsky
    """

    if not param_list:
        param_list = []

    structure: Any = class_object(*param_list)

    struct_len: int = ctypes.sizeof(structure)

    struct_data: bytes | bytearray = buffer[start_offset:start_offset + struct_len]

    least_len: int = min(len(struct_data), struct_len)

    ctypes.memmove(ctypes.addressof(structure), struct_data, least_len)

    return structure



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Copyright (C) 2022-2024 Plato Mavropoulos
"""

import os
import re
import shutil
import sys

from importlib.abc import Loader
from importlib.machinery import ModuleSpec
from importlib.util import module_from_spec, spec_from_file_location
from types import ModuleType
from typing import Type

from biosutilities.common.paths import is_dir, is_file_read, project_root
from biosutilities.common.texts import to_string


def get_external_path(cmd: str | list | tuple) -> str:
    """ Get external dependency path (PATH environment variable or "external" directory) """

    external_root: str = os.path.join(project_root(), 'external')

    external_path: str | None = external_root if is_dir(in_path=external_root) else None

    for command in cmd if isinstance(cmd, (list, tuple)) else [to_string(in_object=cmd)]:
        command_path: str | None = shutil.which(command, path=external_path)

        if command_path and is_file_read(in_path=command_path):
            return command_path

    raise OSError(f'{to_string(in_object=cmd, sep_char=", ")} requirement could not be found!')


def big_script_tool() -> Type | None:
    """ Get Intel BIOS Guard Script Tool class """

    try:
        bgst: str = get_external_path(cmd='big_script_tool')

        bgst_spec: ModuleSpec | None = spec_from_file_location('big_script_tool', re.sub(r'\.PY$', '.py', bgst))

        if bgst_spec and isinstance(bgst_spec.loader, Loader):
            bgst_module: ModuleType | None = module_from_spec(bgst_spec)

            if bgst_module:
                sys.modules['big_script_tool'] = bgst_module

                bgst_spec.loader.exec_module(bgst_module)

                return getattr(bgst_module, 'BigScript')
    except OSError:
        pass

    return None


def comextract_path() -> str:
    """ Get ToshibaComExtractor path """

    return get_external_path(cmd=['comextract', 'ComExtract'])


def szip_path() -> str:
    """ Get 7-Zip path """

    return get_external_path(cmd=['7zzs', '7zz', '7z'])


def tiano_path() -> str:
    """ Get TianoCompress path """

    return get_external_path(cmd=['TianoCompress', 'tianocompress'])


def uefifind_path() -> str:
    """ Get UEFIFind path """

    return get_external_path(cmd=['uefifind', 'UEFIFind'])


def uefiextract_path() -> str:
    """ Get UEFIExtract path """

    return get_external_path(cmd=['uefiextract', 'UEFIExtract'])
